"""Data Fabric tool module for entity-based SQL queries."""

from .datafabric_tool import (
    create_datafabric_query_tool,
    fetch_entity_schemas,
)

__all__ = [
    "create_datafabric_query_tool",
    "fetch_entity_schemas",
]
